@@ -36,8 +36,41 @@ impl EnvEntry {
     }
 }
 
+#[cfg(unix)]
+fn get_shell() -> String {
+    use nix::unistd::{access, AccessFlags};
+    use std::ffi::CStr;
+    use std::path::Path;
+    use std::str;
+
+    let ent = unsafe { libc::getpwuid(libc::getuid()) };
+    if !ent.is_null() {
+        let shell = unsafe { CStr::from_ptr((*ent).pw_shell) };
+        match shell.to_str().map(str::to_owned) {
+            Err(err) => {
+                log::warn!(
+                    "passwd database shell could not be \
+                     represented as utf-8: {err:#}, \
+                     falling back to /bin/sh"
+                );
+            }
+            Ok(shell) => {
+                if let Err(err) = access(Path::new(&shell), AccessFlags::X_OK) {
+                    log::warn!(
+                        "passwd database shell={shell:?} which is \
+                         not executable ({err:#}), falling back to /bin/sh"
+                    );
+                } else {
+                    return shell;
+                }
+            }
+        }
+    }
+    "/bin/sh".into()
+}
+
 fn get_base_env() -> BTreeMap<OsString, EnvEntry> {
-    std::env::vars_os()
+    let mut env: BTreeMap<OsString, EnvEntry> = std::env::vars_os()
         .map(|(key, value)| {
             (
                 EnvEntry::map_key(key.clone()),
@@ -48,7 +81,113 @@ fn get_base_env() -> BTreeMap<OsString, EnvEntry> {
                 },
             )
         })
-        .collect()
+        .collect();
+
+    #[cfg(unix)]
+    {
+        env.insert(
+            EnvEntry::map_key("SHELL".into()),
+            EnvEntry {
+                is_from_base_env: true,
+                preferred_key: "SHELL".into(),
+                value: get_shell().into(),
+            },
+        );
+    }
+
+    #[cfg(windows)]
+    {
+        use std::os::windows::ffi::OsStringExt;
+        use winapi::um::processenv::ExpandEnvironmentStringsW;
+        use winreg::enums::{RegType, HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE};
+        use winreg::types::FromRegValue;
+        use winreg::{RegKey, RegValue};
+
+        fn reg_value_to_string(value: &RegValue) -> anyhow::Result<OsString> {
+            match value.vtype {
+                RegType::REG_EXPAND_SZ => {
+                    let src = unsafe {
+                        std::slice::from_raw_parts(
+                            value.bytes.as_ptr() as *const u16,
+                            value.bytes.len() / 2,
+                        )
+                    };
+                    let size =
+                        unsafe { ExpandEnvironmentStringsW(src.as_ptr(), std::ptr::null_mut(), 0) };
+                    let mut buf = vec![0u16; size as usize + 1];
+                    unsafe {
+                        ExpandEnvironmentStringsW(src.as_ptr(), buf.as_mut_ptr(), buf.len() as u32)
+                    };
+
+                    let mut buf = buf.as_slice();
+                    while let Some(0) = buf.last() {
+                        buf = &buf[0..buf.len() - 1];
+                    }
+                    Ok(OsString::from_wide(buf))
+                }
+                _ => Ok(OsString::from_reg_value(value)?),
+            }
+        }
+
+        if let Ok(sys_env) = RegKey::predef(HKEY_LOCAL_MACHINE)
+            .open_subkey("System\\CurrentControlSet\\Control\\Session Manager\\Environment")
+        {
+            for res in sys_env.enum_values() {
+                if let Ok((name, value)) = res {
+                    if name.to_ascii_lowercase() == "username" {
+                        continue;
+                    }
+                    if let Ok(value) = reg_value_to_string(&value) {
+                        log::trace!("adding SYS env: {:?} {:?}", name, value);
+                        env.insert(
+                            EnvEntry::map_key(name.clone().into()),
+                            EnvEntry {
+                                is_from_base_env: true,
+                                preferred_key: name.into(),
+                                value,
+                            },
+                        );
+                    }
+                }
+            }
+        }
+
+        if let Ok(sys_env) = RegKey::predef(HKEY_CURRENT_USER).open_subkey("Environment") {
+            for res in sys_env.enum_values() {
+                if let Ok((name, value)) = res {
+                    if let Ok(value) = reg_value_to_string(&value) {
+                        // Merge the system and user paths together
+                        let value = if name.to_ascii_lowercase() == "path" {
+                            match env.get(&EnvEntry::map_key(name.clone().into())) {
+                                Some(entry) => {
+                                    let mut result = OsString::new();
+                                    result.push(&entry.value);
+                                    result.push(";");
+                                    result.push(&value);
+                                    result
+                                }
+                                None => value,
+                            }
+                        } else {
+                            value
+                        };
+
+                        log::trace!("adding USER env: {:?} {:?}", name, value);
+                        env.insert(
+                            EnvEntry::map_key(name.clone().into()),
+                            EnvEntry {
+                                is_from_base_env: true,
+                                preferred_key: name.into(),
+                                value,
+                            },
+                        );
+                    }
+                }
+            }
+        }
+    }
+
+    env
 }
 
 /// `CommandBuilder` is used to prepare a command to be spawned into a pty.
@@ -61,6 +200,7 @@ pub struct CommandBuilder {
     cwd: Option<OsString>,
     #[cfg(unix)]
     pub(crate) umask: Option<libc::mode_t>,
+    controlling_tty: bool,
 }
 
 impl CommandBuilder {
@@ -73,6 +213,7 @@ impl CommandBuilder {
             cwd: None,
             #[cfg(unix)]
             umask: None,
+            controlling_tty: true,
         }
     }
 
@@ -84,9 +225,24 @@ impl CommandBuilder {
             cwd: None,
             #[cfg(unix)]
             umask: None,
+            controlling_tty: true,
         }
     }
 
+    /// Set whether we should set the pty as the controlling terminal.
+    /// The default is true, which is usually what you want, but you
+    /// may need to set this to false if you are crossing container
+    /// boundaries (eg: flatpak) to workaround issues like:
+    /// <https://github.com/flatpak/flatpak/issues/3697>
+    /// <https://github.com/flatpak/flatpak/issues/3285>
+    pub fn set_controlling_tty(&mut self, controlling_tty: bool) {
+        self.controlling_tty = controlling_tty;
+    }
+
+    pub fn get_controlling_tty(&self) -> bool {
+        self.controlling_tty
+    }
+
     /// Create a new builder instance that will run some idea of a default
     /// program.  Such a builder will panic if `arg` is called on it.
     pub fn new_default_prog() -> Self {
@@ -96,6 +252,7 @@ impl CommandBuilder {
             cwd: None,
             #[cfg(unix)]
             umask: None,
+            controlling_tty: true,
         }
     }
 
@@ -124,6 +281,14 @@ impl CommandBuilder {
         }
     }
 
+    pub fn get_argv(&self) -> &Vec<OsString> {
+        &self.args
+    }
+
+    pub fn get_argv_mut(&mut self) -> &mut Vec<OsString> {
+        &mut self.args
+    }
+
     /// Override the value of an environmental variable
     pub fn env<K, V>(&mut self, key: K, value: V)
     where
@@ -154,7 +319,7 @@ impl CommandBuilder {
         self.envs.clear();
     }
 
-    fn get_env<K>(&self, key: K) -> Option<&OsStr>
+    pub fn get_env<K>(&self, key: K) -> Option<&OsStr>
     where
         K: AsRef<OsStr>,
     {
@@ -175,6 +340,10 @@ impl CommandBuilder {
         self.cwd = Some(dir.as_ref().to_owned());
     }
 
+    pub fn clear_cwd(&mut self) {
+        self.cwd.take();
+    }
+
     pub fn get_cwd(&self) -> Option<&OsString> {
         self.cwd.as_ref()
     }
@@ -200,6 +369,20 @@ impl CommandBuilder {
         )
     }
 
+    pub fn iter_full_env_as_str(&self) -> impl Iterator<Item = (&str, &str)> {
+        self.envs.values().filter_map(
+            |EnvEntry {
+                 preferred_key,
+                 value,
+                 ..
+             }| {
+                let key = preferred_key.to_str()?;
+                let value = value.to_str()?;
+                Some((key, value))
+            },
+        )
+    }
+
     /// Return the configured command and arguments as a single string,
     /// quoted per the unix shell conventions.
     pub fn as_unix_command_line(&self) -> anyhow::Result<String> {
@@ -225,7 +408,9 @@ impl CommandBuilder {
     }
 
     fn search_path(&self, exe: &OsStr, cwd: &OsStr) -> anyhow::Result<OsString> {
+        use nix::unistd::{access, AccessFlags};
         use std::path::Path;
+
         let exe_path: &Path = exe.as_ref();
         if exe_path.is_relative() {
             let cwd: &Path = cwd.as_ref();
@@ -237,7 +422,7 @@ impl CommandBuilder {
             if let Some(path) = self.resolve_path() {
                 for path in std::env::split_paths(&path) {
                     let candidate = path.join(&exe);
-                    if candidate.exists() {
+                    if access(&candidate, AccessFlags::X_OK).is_ok() {
                         return Ok(candidate.into_os_string());
                     }
                 }
@@ -248,9 +433,10 @@ impl CommandBuilder {
                 exe_path.display()
             );
         } else {
-            if !exe_path.exists() {
+            if let Err(err) = access(exe_path, AccessFlags::X_OK) {
                 anyhow::bail!(
-                    "Unable to spawn {} because it doesn't exist on the filesystem",
+                    "Unable to spawn {} because it doesn't exist on the filesystem \
+                    or is not executable ({err:#})",
                     exe_path.display()
                 );
             }
@@ -270,10 +456,9 @@ impl CommandBuilder {
             .map(|dir| dir.as_os_str())
             .filter(|dir| std::path::Path::new(dir).is_dir())
             .unwrap_or(home.as_ref());
+        let shell = self.get_shell();
 
         let mut cmd = if self.is_default_prog() {
-            let shell = self.get_shell()?;
-
             let mut cmd = std::process::Command::new(&shell);
 
             // Run the shell as a login shell by prefixing the shell's
@@ -292,6 +477,7 @@ impl CommandBuilder {
         cmd.current_dir(dir);
 
         cmd.env_clear();
+        cmd.env("SHELL", shell);
         cmd.envs(self.envs.values().map(
             |EnvEntry {
                  is_from_base_env: _,
@@ -306,23 +492,20 @@ impl CommandBuilder {
     /// Determine which shell to run.
     /// We take the contents of the $SHELL env var first, then
     /// fall back to looking it up from the password database.
-    fn get_shell(&self) -> anyhow::Result<String> {
+    pub fn get_shell(&self) -> String {
+        use nix::unistd::{access, AccessFlags};
+
         if let Some(shell) = self.get_env("SHELL").and_then(OsStr::to_str) {
-            return Ok(shell.into());
+            match access(shell, AccessFlags::X_OK) {
+                Ok(()) => return shell.into(),
+                Err(err) => log::warn!(
+                    "$SHELL -> {shell:?} which is \
+                     not executable ({err:#}), falling back to password db lookup"
+                ),
+            }
         }
 
-        let ent = unsafe { libc::getpwuid(libc::getuid()) };
-        if ent.is_null() {
-            Ok("/bin/sh".into())
-        } else {
-            use std::ffi::CStr;
-            use std::str;
-            let shell = unsafe { CStr::from_ptr((*ent).pw_shell) };
-            shell
-                .to_str()
-                .map(str::to_owned)
-                .context("failed to resolve shell")
-        }
+        get_shell().into()
     }
 
     fn get_home_dir(&self) -> anyhow::Result<String> {
@@ -426,6 +609,15 @@ impl CommandBuilder {
         block
     }
 
+    pub fn get_shell(&self) -> String {
+        let exe: OsString = self
+            .get_env("ComSpec")
+            .unwrap_or(OsStr::new("cmd.exe"))
+            .into();
+        exe.into_string()
+            .unwrap_or_else(|_| "%CompSpec%".to_string())
+    }
+
     pub(crate) fn cmdline(&self) -> anyhow::Result<(Vec<u16>, Vec<u16>)> {
         let mut cmdline = Vec::<u16>::new();
 
