@@ -4,12 +4,13 @@ use crate::{Child, CommandBuilder, MasterPty, PtyPair, PtySize, PtySystem, Slave
 use anyhow::{bail, Error};
 use filedescriptor::FileDescriptor;
 use libc::{self, winsize};
-use std::io;
+use std::cell::RefCell;
 use std::io::{Read, Write};
-use std::mem;
-use std::os::unix::io::{AsRawFd, FromRawFd, RawFd};
+use std::os::unix::io::{AsRawFd, FromRawFd};
 use std::os::unix::process::CommandExt;
-use std::ptr;
+use std::{io, mem, ptr};
+
+pub use std::os::unix::io::RawFd;
 
 #[derive(Default)]
 pub struct UnixPtySystem {}
@@ -43,6 +44,7 @@ fn openpty(size: PtySize) -> anyhow::Result<(UnixMasterPty, UnixSlavePty)> {
 
     let master = UnixMasterPty {
         fd: PtyFd(unsafe { FileDescriptor::from_raw_fd(master) }),
+        took_writer: RefCell::new(false),
     };
     let slave = UnixSlavePty {
         fd: PtyFd(unsafe { FileDescriptor::from_raw_fd(slave) }),
@@ -193,6 +195,7 @@ impl PtyFd {
         let configured_umask = builder.umask;
 
         let mut cmd = builder.as_command()?;
+        let controlling_tty = builder.get_controlling_tty();
 
         unsafe {
             cmd.stdin(self.as_stdio()?)
@@ -223,7 +226,7 @@ impl PtyFd {
                     // are system dependent, which is why we're using `as _`.
                     // Suppress this lint for this section of code.
                     #[cfg_attr(feature = "cargo-clippy", allow(clippy::cast_lossless))]
-                    {
+                    if controlling_tty {
                         // Set the pty as the controlling terminal.
                         // Failure to do this means that delivery of
                         // SIGWINCH won't happen when we resize the
@@ -262,6 +265,7 @@ impl PtyFd {
 /// The file descriptor will be closed when the Pty is dropped.
 struct UnixMasterPty {
     fd: PtyFd,
+    took_writer: RefCell<bool>,
 }
 
 /// Represents the slave end of a pty.
@@ -312,9 +316,17 @@ impl MasterPty for UnixMasterPty {
         Ok(Box::new(fd))
     }
 
-    fn try_clone_writer(&self) -> Result<Box<dyn Write + Send>, Error> {
+    fn take_writer(&self) -> Result<Box<dyn Write + Send>, Error> {
+        if *self.took_writer.borrow() {
+            anyhow::bail!("cannot take writer more than once");
+        }
+        *self.took_writer.borrow_mut() = true;
         let fd = PtyFd(self.fd.try_clone()?);
-        Ok(Box::new(UnixMasterPty { fd }))
+        Ok(Box::new(UnixMasterWriter { fd }))
+    }
+
+    fn as_raw_fd(&self) -> Option<RawFd> {
+        Some(self.fd.0.as_raw_fd())
     }
 
     fn process_group_leader(&self) -> Option<libc::pid_t> {
@@ -323,9 +335,34 @@ impl MasterPty for UnixMasterPty {
             _ => None,
         }
     }
+
+    fn get_termios(&self) -> Option<nix::sys::termios::Termios> {
+        nix::sys::termios::tcgetattr(self.fd.0.as_raw_fd()).ok()
+    }
 }
 
-impl Write for UnixMasterPty {
+/// Represents the master end of a pty.
+/// EOT will be sent, and then the file descriptor will be closed when
+/// the Pty is dropped.
+struct UnixMasterWriter {
+    fd: PtyFd,
+}
+
+impl Drop for UnixMasterWriter {
+    fn drop(&mut self) {
+        let mut t: libc::termios = unsafe { std::mem::MaybeUninit::zeroed().assume_init() };
+        if unsafe { libc::tcgetattr(self.fd.0.as_raw_fd(), &mut t) } == 0 {
+            // EOF is only interpreted after a newline, so if it is set,
+            // we send a newline followed by EOF.
+            let eot = t.c_cc[libc::VEOF];
+            if eot != 0 {
+                let _ = self.fd.0.write_all(&[b'\n', eot]);
+            }
+        }
+    }
+}
+
+impl Write for UnixMasterWriter {
     fn write(&mut self, buf: &[u8]) -> Result<usize, io::Error> {
         self.fd.write(buf)
     }
